﻿using System;
using System.Collections.Generic;
using System.Linq;
using Clarion;
using Clarion.Framework;
using Clarion.Framework.Core;
using Clarion.Framework.Templates;
using WorldServerLibrary.Model;
using System.Threading;

namespace ClarionSimulation
{
    /// <summary>

    /// This event will be fired at the beginning of each cognitive cycle in order to get the current visual sensor information.
    /// </summary>
    /// <returns>The Visual Sensor Information Itens</returns> 
    public delegate IList<Thing> InputVisualSensorialInformationEventHandler();

    /// <summary>
    /// This event will be fired at the end of each cognitive cycle in order to translate the selected action to a World Server Command
    /// </summary>
    /// <param name="externalAction">The external action that was choice</param>
    public delegate void OutputActionChunkEventHandler(ClarionAgentActionType externalAction);

    /// <summary>
    /// Public enum that represents all possibilities of agent actions
    /// </summary>
    public enum ClarionAgentActionType
    {	
		EATFOOD,
		TAKEJEWEL,
        //DO_NOTHING,
        ROTATE_CLOCKWISE,
        ROTATE,
		MOVE_FOOD,
		MOVE_JEWEL
    }

    public class ClarionAgent
    {
        #region Constants

        /// <summary>
        /// Constant that represents the Visual Sensor
        /// </summary>
        private String SENSOR_VISUAL_DIMENSION = "VisualSensor";

        /// <summary>
        /// Constant that represents that there is at least one wall ahead
        /// </summary>
        //private String DIMENSION_WALL_AHEAD = "WallAhead";
		private String DIMENSION_MOVE_FOOD = "Move_food";
		private String DIMENSION_MOVE_JEWEL = "Move_jewel";
		private String DIMENSION_EATFOOD = "EatFood";
		private String DIMENSION_TAKEJEWEL = "TakeJewel";
		private String DIMENSION_ROTATE = "Rotate";

        #endregion

        #region Properties

        #region Simulation

        /// <summary>
        /// If this value is greater than zero, the agent will have a finite number of cognitive cycle. Otherwise, it will have infinite cycles.
        /// </summary>
        public double MaxNumberOfCognitiveCycles = -1;

        /// <summary>
        /// Current cognitive cycle number
        /// </summary>
        private double CurrentCognitiveCycle = 0;

        /// <summary>
        /// Time between cognitive cycle in miliseconds
        /// </summary>
        public Int32 TimeBetweenCognitiveCycles = 0;

        /// <summary>
        /// A thread Class that will handle the simulation process
        /// </summary>
        private Thread runThread;

        #endregion

        #region Agent

        /// <summary>
        /// The agent 
        /// </summary>
        private Clarion.Framework.Agent CurrentAgent;

        /// <summary>
        /// For each cognitive cycle, this event will be called in order to the agent receives the current sensorial information
        /// </summary>
        public event InputVisualSensorialInformationEventHandler OnNewVisualSensorialInformation;

        /// <summary>
        /// For each cognitive cycle, this event will be called when the agent selects one action
        /// </summary>
        public event OutputActionChunkEventHandler OnNewExternalActionSelected;

        #endregion

        #region Perception Input

        /// <summary>
        /// Perception input to indicates a wall ahead
        /// </summary>
        //private DimensionValuePair InputWallAhead;
		private DimensionValuePair InputEatFood;
		private DimensionValuePair InputTakejewel;
		private DimensionValuePair InputMove_food;
		private DimensionValuePair InputMove_jewel;
		private DimensionValuePair InputRotate;

        #endregion

        #region Action Output

        /// <summary>
        /// Output action that makes the agent to rotate clockwise
        /// </summary>
        //private ExternalActionChunk OutputRotateClockwise;
		private ExternalActionChunk OutputEatfood;
		private ExternalActionChunk OutputMove_food;
		private ExternalActionChunk OutputMove_jewel;
		private ExternalActionChunk OutputTakejewel;
        /// <summary>
        /// Output action that makes the agent go ahead
        /// </summary>
        private ExternalActionChunk OutputRotate;

        #endregion

        #endregion

        #region Constructor

        public ClarionAgent()
        {
            // Initialize the agent
            CurrentAgent = World.NewAgent("Current Agent");

            // Initialize Input Information
            //InputWallAhead = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_WALL_AHEAD);
			InputEatFood= World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_EATFOOD);
			InputMove_food= World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_MOVE_FOOD);
			InputMove_jewel= World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_MOVE_JEWEL);
			InputRotate= World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_ROTATE);
			InputTakejewel= World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_TAKEJEWEL);
            // Initialize Output actions
            //OutputRotateClockwise = World.NewExternalActionChunk(ClarionAgentActionType.ROTATE_CLOCKWISE.ToString());
            OutputRotate = World.NewExternalActionChunk(ClarionAgentActionType.ROTATE.ToString());
			OutputEatfood = World.NewExternalActionChunk(ClarionAgentActionType.EATFOOD.ToString());
			OutputTakejewel = World.NewExternalActionChunk(ClarionAgentActionType.TAKEJEWEL.ToString());
			OutputMove_food = World.NewExternalActionChunk(ClarionAgentActionType.MOVE_FOOD.ToString());
			OutputMove_jewel = World.NewExternalActionChunk(ClarionAgentActionType.MOVE_JEWEL.ToString());
            //Create thread to simulation
            runThread = new Thread(RunThread);
			Console.WriteLine("Agent started");
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Run the Simulation in World Server 3d Environment
        /// </summary>
        public void Run()
        {                
			Console.WriteLine ("Running ...");
            // Setup Agent to run
            if (runThread != null && !runThread.IsAlive)
            {
                SetupAgentInfraStructure();

                if (OnNewVisualSensorialInformation != null)
                {
                    // Start Simulation Thread                
                    runThread.Start(null);
                }
            }
        }

        /// <summary>
        /// Abort the current Simulation
        /// </summary>
        /// <param name="deleteAgent">If true beyond abort the current simulation it will die the agent.</param>
        public void Abort(Boolean deleteAgent)
        {   Console.WriteLine ("Aborting ...");
            if (runThread != null && runThread.IsAlive)
            {
                runThread.Abort();
            }

            if (CurrentAgent != null && deleteAgent)
            {
                CurrentAgent.Die();
            }
        }

        #endregion

        #region Setup Agent Methods

        /// <summary>
        /// Setup agent infra structure (ACS, NACS, MS and MCS)
        /// </summary>
        private void SetupAgentInfraStructure()
        {
            // Setup the ACS Subsystem
            SetupACS();                    
        }

        private void SetupMS()
        {            
            //RichDrive
        }

        /// <summary>
        /// Setup the ACS subsystem
        /// </summary>
        private void SetupACS()
        {
			SupportCalculator moveF = FixedRuleDelegateToMoveFood;
            FixedRule ruleAvoidMove = 
				AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputMove_food, moveF);
            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit(ruleAvoidMove);

			SupportCalculator moveJ = FixedRuleDelegateToMoveJewel;
			FixedRule ruleAvoidMoveJewel = 
				AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputMove_jewel, moveJ);
			// Commit this rule to Agent (in the ACS)
			CurrentAgent.Commit(ruleAvoidMoveJewel);

            SupportCalculator eatfoodSupportCalculator = FixedRuleDelegateToEatFood;
            FixedRule ruleAvoidEatfood = 
				AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputEatfood, eatfoodSupportCalculator);
            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit(ruleAvoidEatfood);

			SupportCalculator takeJewelSupportCalculator = FixedRuleDelegateToTakeJewel;
			FixedRule ruleAvoidTakeJewel = 
				AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputTakejewel, takeJewelSupportCalculator);			
			// Commit this rule to Agent (in the ACS)
			CurrentAgent.Commit(ruleAvoidTakeJewel);
            
			SupportCalculator rotateSupportCalculator = FixedRuleDelegateToRotate;
            FixedRule ruleRotate = 
				AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputRotate, rotateSupportCalculator);
            CurrentAgent.Commit(ruleRotate);

            // Disable Rule Refinement
            CurrentAgent.ACS.Parameters.PERFORM_RER_REFINEMENT = false;

            // The selectio type will be probabilistic
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_METHOD = ActionCenteredSubsystem.LevelSelectionMethods.STOCHASTIC;

            // The selection of the action will be fixed (not variable) i.e. only the statement defined above.
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_OPTION = ActionCenteredSubsystem.LevelSelectionOptions.FIXED;

            // Define Probabilistic values
            CurrentAgent.ACS.Parameters.FIXED_FR_LEVEL_SELECTION_MEASURE = 1;
            CurrentAgent.ACS.Parameters.FIXED_IRL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_BL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_RER_LEVEL_SELECTION_MEASURE = 0;
        }

		SensoryInformation buscaTudo (IList<Thing> sensorialInformation, SensoryInformation si)
		{
			if (sensorialInformation.Where (item => ((item.CategoryId == Thing.CATEGORY_JEWEL) 
			                                         && item.DistanceToCreature <= 25.0)).Any ()) {
				si.Add (InputTakejewel, 1.0);
				si.Add (InputEatFood, 0.0);
				si.Add (InputMove_food, 0.0);
				si.Add (InputMove_jewel, 0.0);
				si.Add (InputRotate, 0.0);
			}
			else {
				si.Add (InputTakejewel, 0.0);
				if (sensorialInformation.Where (item => ((item.CategoryId == Thing.CATEGORY_FOOD 
				                                          || item.CategoryId == Thing.CATEGORY_NPFOOD 
				                                          || item.CategoryId == Thing.categoryPFOOD) 
				                                         && item.DistanceToCreature <= 25.0)).Any ()) {
					si.Add (InputEatFood, 1.0);
					si.Add (InputMove_food, 0.0);
					si.Add (InputMove_jewel, 0.0);
					si.Add (InputRotate, 0.0);
				}
				else {
					si.Add (InputEatFood, 0.0);
					if (sensorialInformation.Where (item => ((item.CategoryId == Thing.CATEGORY_JEWEL) 
					                                         && item.DistanceToCreature > 25.0)).Any ()) {
						si.Add (InputMove_food, 0.0);
						si.Add (InputMove_jewel, 1.0);
						si.Add (InputRotate, 0.0);
					}
					else {
						si.Add (InputMove_jewel, 0.0);
						if (sensorialInformation.Where (item => ((item.CategoryId == Thing.CATEGORY_FOOD 
						                                          || item.CategoryId == Thing.CATEGORY_NPFOOD 
						                                          || item.CategoryId == Thing.categoryPFOOD) 
						                                         && item.DistanceToCreature > 25.0)).Any ()) {
							si.Add (InputMove_food, 1.0);
							si.Add (InputRotate, 0.0);
						}
						else {
							si.Add (InputMove_food, 0.0);
							si.Add (InputRotate, 1.0);
						}
					}
				}
			}
			return si;
		}

        /// <summary>
        /// Make the agent perception. In other words, translate the information that came from sensors to a new type that the agent can understand
        /// </summary>
        /// <param name="sensorialInformation">The information that came from server</param>
        /// <returns>The perceived information</returns>
        private SensoryInformation MakePerceptionFromSensorialInput (IList<Thing> sensorialInformation)
		{
			// New sensory information
			SensoryInformation si = World.NewSensoryInformation (CurrentAgent);

			// Detect if we have a wall ahead
//			Boolean wallAhead = sensorialInformation.Where (item => (item.CategoryId == Thing.CATEGORY_BRICK && item.DistanceToCreature <= 61)).Any ();
///			double wallAheadActivationValue = wallAhead ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION;
//			si.Add (InputWallAhead, wallAheadActivationValue);
			Creature creature = null;
			foreach (Thing item in sensorialInformation) {
				if (item.CategoryId == Thing.CATEGORY_CREATURE) {
					creature = (Creature)item;
				}
			}
			if (sensorialInformation.Where (item => ((item.CategoryId == Thing.CATEGORY_JEWEL) 
			                                         && item.DistanceToCreature <= 30.0)).Any ()) {
				si.Add (InputTakejewel, 1.0);
				si.Add (InputEatFood, 0.0);
				si.Add (InputMove_food, 0.0);
				si.Add (InputMove_jewel, 0.0);
				si.Add (InputRotate, 0.0);
			} else if (sensorialInformation.Where (item => ((item.CategoryId == Thing.CATEGORY_FOOD 
			                                                 || item.CategoryId == Thing.CATEGORY_NPFOOD 
			                                                 || item.CategoryId == Thing.categoryPFOOD) 
			                                                && item.DistanceToCreature <= 30.0)).Any ()) {
					si.Add (InputTakejewel, 0.0);
					si.Add (InputEatFood, 1.0);
					si.Add (InputMove_food, 0.0);
					si.Add (InputMove_jewel, 0.0);
					si.Add (InputRotate, 0.0);
				} else if (creature != null && creature.cumpriuLeaflets()) {
				int numLeaflets=creature.NumberOfLeaflets;
				Objetivo obj=null;
				int recompensa=int.MinValue;
				int posicaoMaiorRecompensa=0;
				for (int i = 0; i < numLeaflets; i++){
					obj=creature.Objetivos[i];
					if(!obj.objetivoCumprido() && recompensa<obj.recompensa){
						recompensa=obj.recompensa;
						posicaoMaiorRecompensa=i;
					}
				}
				obj=creature.Objetivos[posicaoMaiorRecompensa];
				//List<Leaflet> procurados=new List<Leaflet>();
				List<String> procurados=new List<string>();
				foreach (Leaflet leaflet in obj.leaflets) {
					if(leaflet.continuarProcurando()){
						procurados.Add(leaflet.Color);
					}
				}
				List<Thing> jewelNoMundo=new List<Thing>();
				foreach (Thing coisa in sensorialInformation) {
					if(coisa.CategoryId==Thing.CATEGORY_JEWEL 
					   && procurados.Contains(coisa.Material.Color)){
						jewelNoMundo.Add(coisa);
					}
				}
				if(jewelNoMundo.Count>0){
					Thing jewelMP=jewelNoMundo[0];
					foreach (Thing joia in jewelNoMundo) {
						if(jewelMP.DistanceToCreature >= joia.DistanceToCreature){
							jewelMP=joia;
						}
					}
					if(jewelMP.DistanceToCreature <=25.0){
						si.Add (InputTakejewel, 1.0);
						si.Add (InputEatFood, 0.0);
						si.Add (InputMove_food, 0.0);
						si.Add (InputMove_jewel, 0.0);
						si.Add (InputRotate, 0.0);
					}else{
						si.Add (InputTakejewel, 0.0);
						si.Add (InputEatFood, 0.0);
						si.Add (InputMove_food, 0.0);
						si.Add (InputMove_jewel, 1.0);
						si.Add (InputRotate, 0.0);
					}
				}else{
					si= buscaTudo (sensorialInformation, si);
				}
			} else {
				si= buscaTudo (sensorialInformation, si);
			}	
            return si;
        }

        #endregion

        #region Delegate Methods

        #region Fixed Rules

        //private double FixedRuleDelegateToAvoidColissionWall(ActivationCollection currentInput, Rule target)
        //{
            // See partial match threshold to verify what are the rules available for action selection
        //    return ((currentInput.Contains(InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
        //}

        private double FixedRuleDelegateToEatFood(ActivationCollection currentInput, Rule target)
        {
			if((currentInput.Contains(InputEatFood, CurrentAgent.Parameters.MAX_ACTIVATION))){
				return 1.0;
			}else{
				return 0.0;
			}            
        }
		private double FixedRuleDelegateToTakeJewel(ActivationCollection currentInput, Rule target)
		{
			if((currentInput.Contains(InputTakejewel, CurrentAgent.Parameters.MAX_ACTIVATION))){
				return 1.0;
			}else{
				return 0.0;
			}
		}

		private double FixedRuleDelegateToMoveFood(ActivationCollection currentInput, Rule target)
        {         
			return ((currentInput.Contains(InputMove_jewel, CurrentAgent.Parameters.MIN_ACTIVATION))
			        &&(currentInput.Contains(InputMove_food, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
        }

		private double FixedRuleDelegateToMoveJewel(ActivationCollection currentInput, Rule target)
		{
			return ((currentInput.Contains(InputMove_jewel, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
		}

		private double FixedRuleDelegateToRotate(ActivationCollection currentInput, Rule target)
        {
            if(currentInput.Contains(InputEatFood, CurrentAgent.Parameters.MIN_ACTIVATION)
			   && (currentInput.Contains(InputTakejewel, CurrentAgent.Parameters.MIN_ACTIVATION))
			   && (currentInput.Contains(InputMove_food, CurrentAgent.Parameters.MIN_ACTIVATION))
			    && (currentInput.Contains(InputMove_jewel, CurrentAgent.Parameters.MIN_ACTIVATION)))			   
			   {
				return 1.0;
			}else{ 
				return 0.0;
			}
        }
        #endregion

        #region Run Thread Method

        private void RunThread(object obj)
        {
			Console.WriteLine("Starting Cognitive Cycle ... press CTRL-C to finish !");
            // Cognitive Cycle starts here getting sensorial information
            while (CurrentCognitiveCycle != MaxNumberOfCognitiveCycles)
            {   
                // Get current sensorial information                    
                IList<Thing> sensorialInformation = OnNewVisualSensorialInformation();

                // Make the perception
                SensoryInformation si = MakePerceptionFromSensorialInput(sensorialInformation);

                //Perceive the sensory information
                CurrentAgent.Perceive(si);

                //Choose an action
                ExternalActionChunk chosen = CurrentAgent.GetChosenExternalAction(si);

                // Get the selected action
                String actionLabel = chosen.LabelAsIComparable.ToString();
				Console.WriteLine(actionLabel);
                ClarionAgentActionType actionType = (ClarionAgentActionType)Enum.Parse(typeof(ClarionAgentActionType), actionLabel, true);

                // Call the output event handler
                if (OnNewExternalActionSelected != null)
                {
                    OnNewExternalActionSelected(actionType);
                }

                // Increment the number of cognitive cycles
                CurrentCognitiveCycle++;

                //Wait to the agent accomplish his job
                if (TimeBetweenCognitiveCycles > 0)
                {
                    Thread.Sleep(TimeBetweenCognitiveCycles);
                }
            }
        }

        #endregion

        #endregion
    }
}
